//! @brief calculate on-the-fly statistics
//! @copyright (c) since 2024, see LICENSE.txt

#include "otf_stats.h"

#include <algorithm>
#include <cmath>
#include <iomanip>
#include <sstream>

// --------------------
OTFStats::OTFStats(std::string tag)
    : m_tag(std::move(tag))
{
    init();
}

// --------------------
void OTFStats::init()
{
    m_num_elems = 0;
    m_min = std::nan("nan");
    m_max = std::nan("nan");

    m_mean = std::nan("nan");
    m_mean2 = 0.0;
    m_variance = std::nan("nan");
    m_stddev = std::nan("nan");
}

// --------------------
auto OTFStats::tag() const -> std::string
{
    return m_tag;
}

// --------------------
void OTFStats::update(double val)
{
    ++m_num_elems;

    // calc min/max
    if (std::isnan(m_min)) {
        m_min = val;
    } else {
        m_min = std::min(val, m_min);
    }
    if (std::isnan(m_max)) {
        m_max = val;
    } else {
        m_max = std::max(val, m_max);
    }

    // calc stddev and average
    double delta = 0.0;
    if (std::isnan(m_mean)) {
        delta = val;
        m_mean = val;
    } else {
        delta = val - m_mean;
        m_mean += delta / (double)m_num_elems;
    }
    const double delta2 = val - m_mean;
    m_mean2 += delta * delta2;
    if (m_num_elems < 2) {
        m_variance = std::nan("nan");
        m_stddev = std::nan("nan");
    } else {
        m_variance = m_mean2 / (double)(m_num_elems - 1);
        m_stddev = std::sqrt(m_variance);
    }
}

// --------------------
void OTFStats::report(writeln_func writeln, bool do_header) const
{
    if (do_header) {
        std::ostringstream ss;
        ss << std::left << std::setw(15) << "Name" << " ";
        ss << std::right;
        ss << std::setw(15) << "Count" << " ";
        ss << std::setw(15) << "Min" << " ";
        ss << std::setw(15) << "Max" << " ";
        ss << std::setw(15) << "Average" << " ";
        ss << std::setw(15) << "Std Dev";
        writeln(ss.str());

        ss.str("");
        ss << std::setw(15) << std::setfill('-') << "" << " ";
        ss << std::setw(15) << std::setfill('-') << "" << " ";
        ss << std::setw(15) << std::setfill('-') << "" << " ";
        ss << std::setw(15) << std::setfill('-') << "" << " ";
        ss << std::setw(15) << std::setfill('-') << "" << " ";
        ss << std::setw(15) << std::setfill('-') << "" << " ";
        writeln(ss.str());
    }

    std::ostringstream ss;
    ss << std::setprecision(6) << std::fixed;
    ss << std::left << std::setw(15) << m_tag << " ";
    ss << std::right;
    ss << std::setw(15) << m_num_elems << " ";
    ss << std::setw(15) << m_min << " ";
    ss << std::setw(15) << m_max << " ";
    ss << std::setw(15) << m_mean << " ";
    ss << std::setw(15) << m_stddev;
    writeln(ss.str());
}

// --------------------
auto OTFStats::num_elements() const -> int_fast64_t
{
    return m_num_elems;
}

// --------------------
auto OTFStats::average() const -> double
{
    return m_mean;
}

// --------------------
auto OTFStats::minimum() const -> double
{
    return m_min;
}

// --------------------
auto OTFStats::maximum() const -> double
{
    return m_max;
}

// --------------------
auto OTFStats::stddev() const -> double
{
    return m_stddev;
}

//    # --------------------
//    ## report the statistics held in all items
//    # the format of this report is minimal as possible
//    #
//    # @param headers (optional) if True (default), add header lines in the report
//    # @return None
//    def report_minimal(self, headers=True):
//        if self.min_max:
//            if headers:
//                self._writeln('Min/Max:')
//            for name, info in sorted(self.min_max.items()):
//                if isinstance(info.maximum, float):
//                    min_str = f'{info.minimum: >15.6f}'
//                else:
//                    min_str = f'{info.minimum: >15}'
//
//                if isinstance(info.maximum, float):
//                    max_str = f'{info.maximum: >15.6f}'
//                else:
//                    max_str = f'{info.maximum: >15}'
//                self._writeln(f'{min_str} {max_str} {name}')
//
//        if self.average:
//            if headers:
//                self._writeln('Average:')
//            for name, info in sorted(self.average.items()):
//                self._writeln(f'     {info.average: >15.6f} {name}')
//
//        if self.stddev:
//            if headers:
//                self._writeln('Stddev')
//            for name, info in sorted(self.stddev.items()):
//                self._writeln(f'{info.stddev: 15.6f} {name}')
//
//        if self.counters:
//            if headers:
//                self._writeln('Counters:')
//            for name, info in sorted(self.counters.items()):
//                self._writeln(f'{info.count: >15} {name}')
//
//    # --------------------
//    ## default routine to write the line to stdout
//    #
//    # @param msg  the line to write. If not provided, an empty line is written
//    # @return None
//    def _default_writeln(self, msg=''):
//        print(msg)